package com.unis.hlht.cjlksj.server.redisstream;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.data.domain.Range;
import org.springframework.data.redis.connection.RedisConnection;
import org.springframework.data.redis.connection.stream.*;
import org.springframework.data.redis.core.RedisCallback;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.StreamOperations;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.time.Duration;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.springframework.data.redis.connection.RedisStreamCommands.XClaimOptions.minIdle;

@Slf4j
@Component
public class RedisStreamScheduled {

    /**
     * 分组名
     */
    public static final String GROUP = "basicack_stream";

    /**
     * 队列名
     */
    public static final String NOTICE_UPLOAD_STREAM_KEY = "notice_upload";

    /**
     * 消费者名称
     */
    public static final String NOTICE_UPLOAD_CONSUMER = "BasicAckConsumer";

    @Autowired
    private RedisTemplate<String, Object> redisTemplate;

    /**
     * 每隔5秒钟，扫描一下有没有等待需要消费的消息
     * 处理死信队列，如果发送给消费者1超过1分钟还没有ack，则转发给消费者2，如果超过20秒，并且转发次数为2，进行手动ack。并且记录异常信息
     */
//    @Scheduled(cron = "0/5 * * * * ?")
    public void scanPendingMsg() {
        StreamOperations<String, Object, Object> streamOperations = this.redisTemplate.opsForStream();
        // 获取group中的pending消息信息，本质上就是执行XPENDING指令
        PendingMessagesSummary pendingMessagesSummary = streamOperations.pending(NOTICE_UPLOAD_STREAM_KEY, GROUP);
        // 所有pending消息的数量
        long totalPendingMessages = pendingMessagesSummary.getTotalPendingMessages();
        if (totalPendingMessages == 0) {
            return;
        }
        // 消费组名称
        String groupName = pendingMessagesSummary.getGroupName();
        // pending队列中的最小ID
        String minMessageId = pendingMessagesSummary.minMessageId();
        // pending队列中的最大ID
        String maxMessageId = pendingMessagesSummary.maxMessageId();
        log.info("流：{},消费组：{}，一共有{}条pending消息，最大ID={}，最小ID={}", NOTICE_UPLOAD_STREAM_KEY, groupName, totalPendingMessages, minMessageId, maxMessageId);
        // 获取每个消费者的pending消息数量
        Map<String, Long> pendingMessagesPerConsumer = pendingMessagesSummary.getPendingMessagesPerConsumer();
        Map<String, List<RecordId>> consumerRecordIdMap = new HashMap<>();
        // 遍历每个消费者中的pending消息
        pendingMessagesPerConsumer.entrySet().forEach(entry -> {
            // 待转组的 RecordId
            List<RecordId> list = new ArrayList<>();
            // 消费者
            String consumer = entry.getKey();
            // 消费者的pending消息数量
            long consumerTotalPendingMessages = entry.getValue();
            log.info("消费者：{}，一共有{}条pending消息", consumer, consumerTotalPendingMessages);
            if (consumerTotalPendingMessages > 0) {
                // 读取消费者pending队列的前10条记录，从ID=0的记录开始，一直到ID最大值
                PendingMessages pendingMessages = streamOperations.pending(NOTICE_UPLOAD_STREAM_KEY, Consumer.from(GROUP, consumer), Range.closed("0", "+"), 10);
                // 遍历所有Opending消息的详情
                pendingMessages.forEach(message -> {
                    // 消息的ID
                    RecordId recordId = message.getId();
                    // 消息从消费组中获取，到此刻的时间
                    Duration elapsedTimeSinceLastDelivery = message.getElapsedTimeSinceLastDelivery();
                    // 消息被获取的次数
                    long deliveryCount = message.getTotalDeliveryCount();
                    // 判断是否超过60秒没有消费
                    if (elapsedTimeSinceLastDelivery.getSeconds() > 120) {
                        // 如果消息被消费的次数为1，则进行一次转组，否则手动消费
                        if (1 == deliveryCount) {
                            list.add(recordId);
                        } else {
                            log.info("手动ACK消息,并记录异常，id={}, elapsedTimeSinceLastDelivery={}, deliveryCount={}", recordId, elapsedTimeSinceLastDelivery, deliveryCount);
                            streamOperations.acknowledge(NOTICE_UPLOAD_STREAM_KEY, GROUP, recordId);
                        }
                    }
                });
                if (list.size() > 0) {
                    consumerRecordIdMap.put(consumer, list);
                }
            }
        });
        // 最后将待转组的消息进行转组
        if (!consumerRecordIdMap.isEmpty()) {
            this.changeConsumer(consumerRecordIdMap);
        }

    }

    /**
     * 将消息进行转组
     *
     * @param consumerRecordIdMap
     */
    private void changeConsumer(Map<String, List<RecordId>> consumerRecordIdMap) {
        consumerRecordIdMap.entrySet().forEach(entry -> {
            // 根据当前consumer去获取另外一个consumer
            String oldComsumer = entry.getKey();
            String newConsumer = entry.getKey();
            List<RecordId> recordIds = entry.getValue();
            List<ByteRecord> retVal = this.redisTemplate.execute(new RedisCallback<List<ByteRecord>>() {
                @Override
                public List<ByteRecord> doInRedis(RedisConnection redisConnection) throws DataAccessException {
                    // 相当于执行XCLAIM操作，批量将某一个consumer中的消息转到另外一个consumer中
                    return redisConnection.streamCommands().xClaim(NOTICE_UPLOAD_STREAM_KEY.getBytes(),
                            GROUP, newConsumer, minIdle(Duration.ofSeconds(10)).ids(recordIds));
                }
            });
            for (ByteRecord byteRecord : retVal) {
                log.info("改了消息的消费者：id={}, value={},newConsumer={}", byteRecord.getId(), byteRecord.getValue(), newConsumer);
            }
        });
    }


}